/*
** Module   : Protocol.java
** Abstract : low level driver of the P2J protocol
**
** Copyright (c) 2005-2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 SIY 20050120   @19501 Created initial version.
** 002 SIY 20050207   @19712 Implemented exception delivering scheme.
**                           Also modified stop() to avoid waiting for 
**                           threads.
** 003 SIY 20050211   @19784 Reset stream to exclude backreferencing.
** 004 NVS 20050308   @20250 Using NetSocket class instead of SSLSocket.
**                           The ObjectInput and ObjectOutput streams are
**                           created in NetSocket.
** 005 SIY 20050315   @20385 Organized imports and other minor cleanups.
**                           Added call to stream flush() method (merge
**                           from NVS).
** 006 SIY 20050329   @20546 Minor changes in member access rights to 
**                           improve performance (change is very small). 
** 007 GES 20060211   @24533 Minor cleanup and added to-dos to exception
**                           processing.
** 008 NVS 20060616   @27376 Exception processing modified. No more stack
**                           trace printing.
** 009 GES 20060726   @28199 Added support for bidirectional state
**                           synchronization.
** 010 GES 20060731   @28256 Greatly improved performance by reducing the
**                           overhead of object graph serialization. The
**                           trick is to serialize the object twice! First
**                           the object is flattened via serialization to
**                           a byte[]. Then this byte[] is serialized to
**                           the socket.  This intermediate step is 
**                           significantly faster than serializing direct
**                           to the socket! The reading was modified to
**                           handle the proper restore.
** 011 GES 20060802   @28293 Remove attachments before enqueueing because
**                           REQUEST_SYNC messages tend to reuse the same
**                           message on the return trip, but just swap
**                           addresses. Thus the same state sync data was
**                           sent back when it didn't need to be in the
**                           case where the attachment would have been 
**                           null (no changes to send).
** 012 NVS 20061017   @30464 Implemented ability to restart the session
**                           without exiting the process.
** 013 GES 20061207   @31612 Added logging and isRunning().
** 014 NVS 20070411   @32929 Protocol threads are created as part of a
**                           thread group named "protocol".
** 015 ECF 20071118   @35884 Reworked exception handling in reader and
**                           writer threads.
** 016 GES 20081024   @40328 Eliminated use of the NetSocket/Socket and
**                           instead now the protocol is only dependent
**                           upon Queue APIs. This reduces dependencies
**                           between components and fosters better
**                           encapsulation of the protocol.
** 017 ECF 20090205   @42106 Changed to accommodate updated Queue API.
**                           Queue.stop() method now accepts a boolean
**                           parameter to optionally switch security
**                           context to the session's user's context.
** 018 GES 20090722   @43334 Better thread names to enhance debugging.
** 019 GES 20090723   @43346 Better handling of OOME on reader/writer threads.
**                           Improved control over logging detail. Synchronized
**                           reader/writer thread termination processing to
**                           avoid extra errors occurring do to a race
**                           condition.
** 020 CA  20090902   @43821 Changed logging level for any SocketException's
**                           caught while the queue is still running to always
**                           use FINE.
** 021 CA  20131121          The sync'ed state processing is moved from Reader/Writer threads to 
**                           the thread which actually sends/receives the message (i.e. 
**                           Conversation thread).
*/

package com.goldencode.p2j.net;

import java.io.*;
import java.net.*;
import java.util.logging.*;
import com.goldencode.p2j.util.*;

/**
 * This class implements the low level part of the protocol. It provides two
 * threads (Reader and Writer). Reader thread instantiates messages from the
 * stream and passes to the queue for the further processing. Writer thread
 * stores messages to stream. Also, <code>Protocol</code> is responsible for
 * notifying queue about I/O errors.
 * <p>
 * The reading and writing loops implement the proper callbacks for the
 * {@link StateSynchronizer} implementation of bidirectional state
 * synchronization.  This processing can be activated using the
 * <code>registerSynchronizer</code> method.
 * <p>
 * WARNING: the core loops of the reader and writer threads are controlled
 * by a variable which they read outside of a synchronized block. This means
 * that it is possible (depending on the JVM implementation) to get 
 * inconsistent or corrupt results because it may be read in the middle of
 * being changed. At this time, we are not introducing a lock object since
 * this would cause potential contention to occur between the reader and 
 * writer threads. This is something that needs to be tested to ensure its
 * impact is understood but it DOES need to be fixed.
 */
class Protocol
{
   /** Logger. */
   private static final Logger LOG = 
      LogHelper.getLogger(Protocol.class.getName());
      
   /** Thread group for the protocol threads. */
   private static ThreadGroup protoThreads = new ThreadGroup(
      Thread.currentThread().getThreadGroup(), "protocol");

   /** Flag indicating status for Reader/Writer threads. */
   private volatile boolean running = false;

   /** The queue representing the session with which protocol is working. */
   private Queue queue = null;

   /** Reader thread. */
   private Thread readerThread = null;

   /** Writer thread. */
   private Thread writerThread = null;
   
   /** Synchronizes access to the state synchronizer. */
   private Object slock = new Object();
   
   /** The object used to implement state synchronization if enabled. */
   private StateSynchronizer ssync = null;

   /** Synchronizes termination processing. */
   private Object tlock = new Object();
   
   /**
    * Constructs an instance using the provided queue.
    * 
    * @param    queue
    *           Instance of the <code>Queue</code> to/from which messages
    *           are written/read.
    */
   Protocol(Queue queue)
   {
      this.queue = queue;
   }

   /**
    * Starts Reader and Writer threads as daemon threads.  This occurs so
    * long as:
    * <p>
    * <ul>
    *    <li> The threads are not already started.
    *    <li> The queue was properly initialized.
    *    <li> The low-level network transport is not closed.
    * </ul>
    */
   synchronized void start()
   {
      if (running == true || queue == null || queue.isTransportClosed())
         return;

      running = true;

      Runnable reader = new Reader();
      Runnable writer = new Writer();

      readerThread = new Thread(protoThreads, reader, "Reader");
      readerThread.setDaemon(true);

      writerThread = new Thread(protoThreads, writer, "Writer");
      writerThread.setDaemon(true);

      readerThread.start();
      writerThread.start();
   }

   /**
    * Stops Reader and Writer threads.
    */
   synchronized void stop()
   {
      if (!running)
         return;

      running = false;
      readerThread.interrupt();
      writerThread.interrupt();
      
      queue.closeTransport();
      readerThread = null;
      writerThread = null;
   }
   
   /**
    * Force the reader and writer threads to have more descriptive names
    * to aid debugging. This is only meaningful on a router node.
    */
   synchronized void setNames()
   {
      SessionManager sessMgr = SessionManager.get();
      
      if (sessMgr.isRouter())
      {
         String rname = LogHelper.createThreadName("Reader");
         String wname = LogHelper.createThreadName("Writer");
         
         if (rname != null)
            readerThread.setName(rname);
        
         if (wname != null)
            writerThread.setName(wname);
      }
   }
   
   /**
    * Returns the operational state of this protocol driver.
    *
    * @return   <code>true</code> if the protocol is running (operational)
    *           and <code>false</code> if the session is closed.
    */
   public synchronized boolean isRunning()
   {
      return running;
   }
   
   /**
    * Sets the instance of the state synchronizer the reader and writer
    * threads should use, if and only if there is no other state synchronizer
    * already in use.  This "there can BE only ONE" approach can be changed
    * easily in the future but was implemented to ensure high performance
    * since each loop of the protocol must call the synchronization 
    * primitives.
    *
    * @param    ssync
    *           The synchronizer to register or <code>null</code> if the
    *           currently registered synchronizer should be deregistered.
    */
   void registerSynchronizer(StateSynchronizer ssync)
   {
      synchronized (slock)
      {
         if (this.ssync == null || ssync == null)
         {
            this.ssync = ssync;
         }
      }
   }
   
   /**
    * Checks if there is a registered state synchronizer and if so, it is
    * queried to obtain the latest changes in state that must be sent to the
    * other side of the connection.  If any changes are found they are 
    * packaged into the message and returned, otherwise the original message
    * is returned. If no state synchronizer is registered, the original
    * message is returned.
    *
    * @param    outbound
    *           The message that was dequeued for transmission.
    *
    * @return   The updated message (with attached state) if there is a
    *           registered state synchronizer AND there are state changes
    *           to send.  Otherwise the original message will be returned.
    */
   Message attachChanges(Message outbound)
   {
      Serializable state  = null;
      Message      result = outbound;
      
      synchronized (slock)
      {
         if (ssync != null)
         {
            state = ssync.getChanges();
         }
      }
                     
      if (state != null)
      {
         result = new SyncMessage(outbound, state);
      }
      
      return result;
   }

   /**
    * If there is a registered state synchronizer and if there are any 
    * attached state updates sent from the other side of the connection,
    * these updates are passed to the state synchronizer to be applied to
    * this node's state.
    *
    * @param    inbound
    *           The message that is about to be enqueued for processing.
    *
    * @return   An instance of the message without an attachment or the 
    *           given message (unchanged) if there was no attachment. 
    */
   Message applyChanges(Message inbound)
   {
      if (inbound instanceof SyncMessage)
      {
         SyncMessage  smsg  = (SyncMessage) inbound;
         Serializable state = smsg.getState();
         
         synchronized (slock)
         {
            if (ssync != null)
            {
               ssync.applyChanges(state);
            }
         }
         
         // remove the attachment (this copy constructor knows nothing of
         // the subclass so we will only get a simple message as a result)
         inbound = new Message(inbound);
      }
      
      return inbound;
   }

   /**
    * This helper class implements functionality of Reader thread.
    */
   private class Reader
   implements Runnable
   {
      /**
       * This method instantiates messages from the stream and places into
       * the queue for further processing.
       */
      public void run()
      {
         Exception result = null;

         try
         {
            ObjectInputStream is = queue.getObjectInput();
            
            while (running)
            {
               // read the next message from the stream
               Serializable obj = (Serializable) is.readObject();
               
               // re-inflate the flattened object that came in as a byte[]
               // (this is faster than without the intermediate flattened
               // form)
               obj = byteArrayToObj((byte[]) obj);
               
               // hand the message to the dispatcher for processing
               queue.enqueueInbound(obj);
            }
         }
         
         catch (EOFException exc)
         {
            // this happens when the peer side terminates so this is NOT
            // a form of unexpected failure
         }
         
         catch (InterruptedException exc)
         {
            // an interrupted exception will be caused by the stop() method
            // so we consider this a "clean" way of exiting
         }
         
         catch (SocketException exc)
         {
            // this happens when the socket is closed so this is NOT a form
            // of unexpected failure if the running flag was set to false
            if (running)
            {
               // log that there was an unexpected failure
               if (LOG.isLoggable(Level.FINE))
               {
                  LOG.logp(Level.FINE,
                           "Protocol.Reader.run()",
                           "",
                           LogHelper.generate("Peer disconnected!"),
                           exc);
               }
               
               result = exc;
            }
         }
         
         catch (Throwable thr)
         {
            // unexpected failure
            
            if (LOG.isLoggable(Level.WARNING))
            {
               LOG.logp(Level.WARNING,
                        "Protocol.Reader.run()",
                        "",
                        LogHelper.generate("failure in reading loop"),
                        thr);
            }
            
            result = (thr instanceof Exception) ? (Exception) thr
                                                : new RuntimeException(thr);
         }

         finally
         {
            // only one of the (reader and writer) threads should process
            // termination at any given time; the first one in gets to process
            // and the other will be blocked until the processing is complete;
            synchronized (tlock)
            {
               // the second thread will find that the running flag is cleared
               // so it won't need to and won't try to execute termination
               // processing
               if (running)
               {
                  queue.stop(result, true);
               }
            }
         }
      }
      
      /**
       * Restore ("inflate") an object from an in-memory array of bytes.
       *
       * @param    obj
       *           The flattened object to restore.
       *
       * @return   The inflated object.
       */
      private Serializable byteArrayToObj(byte[] obj)
      throws IOException,
             ClassNotFoundException
      {
         ByteArrayInputStream bais = new ByteArrayInputStream(obj);
         ObjectInputStream    ois  = new ObjectInputStream(bais);
         
         return (Serializable) ois.readObject();
      }
   }

   /**
    * This helper class implements functionality of Writer thread.
    */
   private class Writer
   implements Runnable
   {
      /**
       * This method obtains messages from the Outbound Queue and writes them
       * into output stream.
       */
      public void run()
      {
         Exception result = null;

         try
         {
            ObjectOutputStream os = queue.getObjectOutput();

            while (running)
            {
               // dequeue a message
               Object obj = queue.dequeueOutbound();
               
               // flatten the object to a byte[] before writing it to the
               // socket since this is significantly faster than going
               // without the intermediate byte[]; note that we still use
               // serialization for the final output since the low level
               // Java approach handles reading an entire array as a single
               // entity which is something we want to avoid doing ourselves
               os.writeObject(objToByteArray((Serializable) obj));
               os.flush();
               os.reset();
            }
         }
         
         catch (InterruptedException exc)
         {
            // an interrupted exception will be caused by the stop() method
            // so we consider this a "clean" way of exiting
         }
         
         catch (SocketException exc)
         {
            // this happens when the socket is closed so this is NOT a form
            // of unexpected failure if the running flag was set to false
            if (running)
            {
               // log that there was an unexpected failure
               if (LOG.isLoggable(Level.FINE))
               {
                  LOG.logp(Level.FINE,
                           "Protocol.Writer.run()",
                           "",
                           LogHelper.generate("Peer disconnected!"),
                           exc);
               }
               
               result = exc;
            }
         }
         
         catch (Throwable thr)
         {
            // unexpected failure
            
            if (LOG.isLoggable(Level.WARNING))
            {
               LOG.logp(Level.WARNING,
                        "Protocol.Writer.run()",
                        "",
                        LogHelper.generate("failure in writing loop"),
                        thr);
            }
            
            result = (thr instanceof Exception) ? (Exception) thr
                                                : new RuntimeException(thr);
         }

         finally
         {
            // only one of the (reader and writer) threads should process
            // termination at any given time; the first one in gets to process
            // and the other will be blocked until the processing is complete;
            synchronized (tlock)
            {
               // the second thread will find that the running flag is cleared
               // so it won't need to and won't try to execute termination
               // processing
               if (running)
               {
                  queue.stop(result, true);
               }
            }
         }
      }
      
      /**
       * Serialize the given object into an in-memory array of bytes.
       *
       * @param    obj
       *           The graph of objects to be converted.
       *
       * @return   The "flattened" version.
       */
      private byte[] objToByteArray(Serializable obj)
      throws IOException
      {
         ByteArrayOutputStream baos = new ByteArrayOutputStream(8192);
         ObjectOutputStream    oos  = new ObjectOutputStream(baos);
         oos.writeObject(obj);
         return baos.toByteArray();
      }
   }
}