/*
** Module   : FrameConfig.java
** Abstract : stores frame-specific attributes
**
** Copyright (c) 2005-2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 EVL 20050713   @21692 Creation based on general concept. Necessary
**                           modification to get compiled.
** 002 EVL 20050721   @21801 Adding layout specific constants.
** 003 SIY 20050831   @22504 Refactoring (see @@22394-22409) 
** 004 SIY 20050907   @22637 Fixed use of @inheritDoc tag in javadoc,
**                           moved methods to ContainerConfig.
** 005 SIY 20050909   @22667 Fixed frame centering.
** 006 SIY 20050926   @22840 Minor cleanup.
** 007 NVS 20051118   @23417 Added setCentered().
** 008 NVS 20051121   @23418 Added setColumns(int) method and columns var.
**                           Added KEEP-TAB-ORDER related methods and var.
**                           Added NO-HELP related methods and var.
**                           Added NO-HIDE related methods and var.
**                           Added NO-UNDERLINE related methods and var.
**                           Added NO-VALIDATE related methods and var.
**                           Added SCROLL n related methods and var.
**                           Added TOP_ONLY related methods and var.
**                           Added missing setBox() method.
** 009 SIY 20060128   @24119 Added support code for the frame manager.
** 010 SIY 20060209   @24469 Major refactoring.
** 011 SIY 20060213   @24576 Changes for the DialogBox support.
** 012 GES 20060227   @24768 pageTop/pageBottom MUST default to false.
** 013 SIY 20060607   @26963 Fixed applying of some attributes.
** 014 SIY 20060627   @27657 Changed handling of the 'title' attribute.
** 015 GES 20060802   @28408 Made externalizable for better performance.
**                           Many code standard cleanups.
** 016 SIY 20060825   @28890 Encapsulated and removed unused fields.
** 017 SIY 20070501   @33337 Get rid of P2JFrame.
** 018 CA  20071122   @35951 Added RETAIN n related methods and var.                              
** 019 SIY 20080218   @37115 Removed setTitle() because it is not necessary  
**                           and triggers redundant re-layouts while frame 
**                           layout is recalculated anyway after applying  
**                           config.
**                           Ref: #37107
** 020 SIY 20080228   @37201 Javadoc cleanup.
** 021 EVL 20080326   @37691 Fix for incorrect numeric fill-in alignment
**                           for COLUMN Progress statement without explicit
**                           SIDE-LABELS parameter set. Must be left
**                           aligned in this case.
** 022 CA  20080529   @38434 Save the configured no-hide value, so it can
**                           be restored when the frame scope is entered 
**                           again. Ref: #38431
** 023 CA  20080905   @39746 Fixed FRAME:ROW/COL handling.
** 024 SIY 20080923   @39907 Limit FRAME:ROW/COL error reporting to cases
**                           when frame is moved at run time by application
**                           so initial frame placement remain silent. 
** 025 CA  20080930   @39983 Added sizeChanged flag, used to notify the
**                           client if the frame size was changed. Fixed
**                           WIDTH/HEIGHT-CHARS behavior.
** 026 SIY 20100619          Switching to own implementation of character UI.
** 027 SIY 20100804          Changes in imports, some refactorings.
** 028 SIY 20100831          Changes related to refactoring of FrameHelper.
** 029 SIY 20110112          Changes related to new color handling scheme.
** 030 SIY 20111101          Moved from CHUI package. Minor 
**                           cleanups/refactorings in preparation to GUI client.
** 031 SIY 20111202          Updated imports.
** 032 CA  20131121          If the frame is no longer visible, attempt to delete it.
*/

package com.goldencode.p2j.ui;

import java.io.*;

import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.widget.*;

/**
 * This class is used to hold frame and child frame specific attributes.
 */
public class FrameConfig
extends ContainerConfig
{
   /** On server-side, a reference to the associated frame instance. */
   protected GenericFrame frame = null;

   /** RETAIN n option, where 0 means paging frame */
   private int retain = 0;

   /** SCROLL n option, where 0 means paging frame */
   private int scroll = 0;

   /** The Progress DOWN attribute. */
   private int down = 1;

   /** Mark DialogBox. */
   private boolean dialog = false;

   /** The Progress CAREFUL-PAINT attribute. */
   private boolean carefulPaint = false;

   /** Means whether the frame must be centered within the parent container. */
   private boolean centered = false;

   /** The Progress LABELS attribute. */
   private boolean labels = true;

   /** The Progress MANUAL-HIGHLIGHT attribute. */
   private boolean manualHighlight = false;

   /** The Progress OVERLAY attribute. */
   private boolean overlay = false;

   /** The Progress PAGE-BOTTOM attribute. */
   private boolean pageBottom = false;

   /** The Progress PAGE-TOP attribute. */
   private boolean pageTop = false;

   /** The Progress SIDE-LABELS attribute. */
   private boolean sideLabels = false;

   /** The Progress BOX (R/O) attribute. */
   private boolean box = true;

   /** KEEP-TAB-ORDER option. */
   private boolean keepTabOrder = false;

   /** NO-HELP option. */
   private boolean noHelp = false;

   /** 
    * NO-HIDE option - used to track the NO-HIDE status of a frame at runtime. 
    * This will be set to false when the frame goes out of scope.
    */
   private boolean noHide = false;

   /** 
    * The NO-HIDE option as configured at frame definition - used to restore 
    * the NO-HIDE status when the frame scope is opened again.
    */
   private boolean noHideConfig = false;

   /** NO-UNDERLINE option. */
   private boolean noUnderline = false;

   /** NO-VALIDATE option. */
   private boolean noValidate = false;

   /** ATTR-SPACE attribute. */
   private boolean attr_space = false;

   /** Flag indicating position change. */
   private boolean positionChanged = false;

   /** Flag indicating size change. */
   private boolean sizeChanged = false;

   /**
    * Default constructor (only used in deserialization).
    */
   public FrameConfig()
   {
   }
   
   /**
    * Create empty configuration with all default values.
    * 
    * @param   type
    *          Type of the new configuration.
    * @param   instance
    *          A component to which this instance will be attached.
    */
   public FrameConfig(int type, Widget instance)
   {
      super(type, instance);
      setVisibleDirect(false);
   }
   
   /**
    * Create an instance from other config. References to other widgets are
    * also restored.
    * 
    * @param   instance
    *          Instance to connect.
    * @param   cfg
    *          Instance to copy data from.
    */
   public FrameConfig(Widget instance, ComponentConfig cfg)
   {
      super(instance, cfg);
   }

   /**
    * Set new values from the provided config instance.
    * 
    * @param   config
    *          Instance to take values from.
    */
   @Override
   public void applyConfig(ComponentConfig config)
   {
      if (config instanceof FrameConfig)
      {
         FrameConfig other = (FrameConfig) config;
         setSizeChanged(other.sizeChanged);
      }

      super.applyConfig(config);
      
      if (config instanceof FrameConfig)
      {
         FrameConfig other = (FrameConfig) config;
         
         setDown(other.down);
         setDialog(other.dialog);
         setAttrSpace(other.attr_space);
         setCarefulPaint(other.carefulPaint);
         setLabels(other.labels);
         setManualHighlight(other.manualHighlight);
         setOverlay(other.overlay);
         setPageBottom(other.pageBottom);
         setPageTop(other.pageTop);
         setSideLabels(other.sideLabels);
         setCentered(other.centered);
         setBox(other.box);
         setKeepTabOrder(other.keepTabOrder);
         setNoHelp(other.noHelp);
         setNoHide(other.noHide);
         setNoHideConfig(other.noHide);
         setNoUnderline(other.noUnderline);
         setNoValidate(other.noValidate);
         
         //Do this once again to assure that centering is performed properly.
         //TODO: move this code?
         if (instance instanceof Container)
         {
            ((Container) instance).setSize(other.getWidthChars(), other.getHeightChars());
         }
      }
   }

   /**
    * Sets HEIGHT-CHARS Progress attribute.
    * 
    * @param    height
    *           The integer value for the new HEIGHT-CHARS value.
    */
   @Override
   public void setHeightChars(int height)
   {
      if (instance instanceof Frame && 
          !((Frame) instance).canSetHeightChars(height))
      {
         return;
      }

      super.setHeightChars(height);
   }

   /**
    * Set the WIDTH-CHARS Progress attribute.
    * 
    * @param    width
    *           The integer value for the new WIDTH-CHARS value.
    */
   @Override
   public void setWidthChars(int width)
   {
      if (instance instanceof Frame && 
          !((Frame) instance).canSetWidthChars(width))
      {
         return;
      }
   
      super.setWidthChars(width);
   }

   /**
    * Setting the value of the ROW attribute.  This is a 0-based row position
    * and NOT the 1-based position used in Progress.
    * 
    * @param   row
    *          The new row value for the upper left component corner relative
    *          to it's parent.
    */
   @Override
   public void setRow(int row)
   {
      if (instance instanceof Frame)
      {
         if (row != BaseConfig.INV_COORD)
         {
            ((Frame) instance).resetAutoplacing();
         }
         if (row == BaseConfig.INV_COORD &&
             instance.location().y != BaseConfig.INV_COORD)
         {
            row = instance.location().y;
         }
         
         if (!((Frame) instance).canSetRow(row))
         {
            return;
         }
      }


      super.setRow(row);
   }
   
   /**
    * Setting the value of the ROW attribute, held by <code>row</code> field.
    * 
    * @param   row
    *          The new row value for the upper left component corner relative
    *          to it's parent.
    */
   public void setRowAttribute(int row)
   {
      this.row = row;
   }

   /**
    * Get the value of the ROW attribute, as held by <code>row</code> field.
    * 
    * @return   The row coordinate of the upper left component corner.
    */
   public int getRowAttribute()
   {
      return row;
   }

   /**
    * Setting the column coordinate of the upper left corner of the component
    * relative to the parent container.  This is a 0-based column position
    * and NOT the 1-based position used in Progress.
    * 
    * @param   column
    *          The new column value of upper left component corner.
    */
   @Override
   public void setColumn(int column)
   {
      if (instance instanceof Frame)
      {
         if (column != BaseConfig.INV_COORD)
         {
            ((Frame) instance).resetAutoplacing();
         }
         
         if (column == BaseConfig.INV_COORD &&
             instance.location().x != BaseConfig.INV_COORD)
         {
            column = instance.location().x;
         }

         if (!((Frame) instance).canSetColumn(column))
         {
            return;
         }
      }
      
      super.setColumn(column);
   }

   /**
    * Get current state of the <code>position changed</code> flag.
    * 
    * @return  <code>true</code> if frame position was changed.
    */
   public boolean isPositionChanged()
   {
      return positionChanged;
   }
   
   /**
    * Set state of <code>position changed</code> flag.
    * 
    * @param   changed
    *          New value for the flag.
    */
   public void setPositionChanged(boolean changed)
   {
      positionChanged = changed;
   }

   /**
    * Get the <code>sizeChanged</code> flag.
    * 
    * @return  <code>true</code> if frame size was changed.
    */
   public boolean isSizeChanged()
   {
      return sizeChanged;
   }
   
   /**
    * Set the <code>sizeChanged</code> flag.
    * 
    * @param   changed
    *          <code>true</code> if frame size was changed.
    */
   public void setSizeChanged(boolean changed)
   {
      sizeChanged = changed;
   }

   /**
    * Getting the thickness in the units of char for the peer frame border
    * at the bottom of the peer frame. Returns 0 in the case when the BOX
    * attribute is false.
    *
    * @return  The value for the BORDER-BOTTOM-CHARS attribute.
    */
   public int getBorderBottomChars()
   {
      return box ? 1 : 0;
   }

   /**
    * Getting the thickness in the units of char for the peer frame border
    * at the left of the peer frame. Returns 0 in the case when the BOX
    * attribute is false.
    *
    * @return  The value for the BORDER-LEFT-CHARS attribute.
    */
   public int getBorderLeftChars()
   {
      return box ? 1 : 0;
   }

   /**
    * Getting the thickness in the units of char for the peer frame border
    * at the right of the peer frame. Returns 0 in the case when the BOX
    * attribute is false.
    *
    * @return  The value for the BORDER-RIGHT-CHARS attribute.
    */
   public int getBorderRightChars()
   {
      return box ? 1 : 0;
   }

   /**
    * Getting the thickness in the units of char for the peer frame border
    * at the top of the peer frame. Returns 0 in the case when the BOX
    * attribute is false.
    *
    * @return  The value for the BORDER-TOP-CHARS attribute.
    */
   public int getBorderTopChars()
   {
      return box ? 1 : 0;
   }

   /**
    * Check whether the peer frame has the border around it.
    *
    * @return  The boolean value for the BOX attribute.
    */
   public boolean isBox()
   {
      return box;
   }

   /**
    * Sets the BOX option for the frame.
    *
    * @param box
    *        <code>true</code> if the frame comes with the box
    */
   public void setBox(boolean box)
   {
      this.box = box;

      if (instance instanceof Frame)
         ((Frame) instance).setBox(box);
   }
   
   /**
    * Set ATTR-SPACE attribute value.
    * 
    * @param   attr_space
    *          New value for the attribute
    */
   public void setAttrSpace(boolean attr_space)
   {
      this.attr_space = attr_space;
   
      if (instance instanceof Frame)
      {
         ((Frame) instance).setAttrSpace(attr_space);
      }
   }

   /**
    * Get current value of the ATTR-SPACE attribute.
    * 
    * @return  Current value of the attribute.
    */
   public boolean isAttrSpace()
   {
      return attr_space;
   }
   
   /**
    * Getting the handle of the cancel button inside the peer frame. Cancel
    * button is the instance of the Button which pressing will be emulated
    * when the user presses the ESC key inside the peer frame.
    * 
    * @return  The instance of the Button which is cancel button for the peer
    *          frame.
    */
   public Button getCancelButton()
   {
      return cancelButton;
   }

   /**
    * Setting the new cancel button handle for the peer frame.
    * 
    * @param   cancelButton
    *          The new Button instance for CANCEL-BUTTON attribute.
    */
   public void setCancelButton(Button cancelButton)
   {
      this.cancelButton = cancelButton;
   }

   /**
    * Check the state of the CAREFUL-PAINT flag for the peer frame. True
    * careful paint means slow and careful painting when there are overlapped
    * widget inside the peer frame.
    * 
    * @return  The boolean value of the CAREFUL-PAINT attribute.
    */
   public boolean isCarefulPaint()
   {
      return carefulPaint;
   }

   /**
    * Setting the new value of the CAREFUL-PAINT flag.
    * 
    * @param   carefulPaint
    *          The new boolean value for the CAREFUL-PAINT attribute.
    */
   public void setCarefulPaint(boolean carefulPaint)
   {
      this.carefulPaint = carefulPaint;
   }

   /**
    * Checks if the peer frame has a label. Must be used only for static frames.
    *
    * @return  The boolean value for the LABELS attribute.
    */
   @Override
   public boolean isLabels()
   {
      return labels;
   }

   /**
    * Setting the LABELS flag.
    *
    * @param   labels
    *          The new boolean value for the LABELS attribute.
    */
   @Override
   public void setLabels(boolean labels)
   {
      this.labels = labels;
   
      if (instance instanceof Frame)
      {
         ((Frame) instance).setLabels(labels);
      }
   }

   /**
    * Checks whether the peer frame must use the custom highlight design for
    * selected component.
    *
    * @return  The boolean value for the MANUAL-HIGHLIGHT attribute.
    */
   public boolean isManualHighlight()
   {
      return manualHighlight;
   }

   /**
    * Setting the new value for the MANUAL-HIGHLIGHT flag.
    *
    * @param   manualHighlight
    *          The new boolean value for the MANUAL-HIGHLIGHT attribute.
    */
   public void setManualHighlight(boolean manualHighlight)
   {
      this.manualHighlight = manualHighlight;
   }

   /**
    * Check whether the peer frame can overlap any other frame except the frame
    * with TOP-ONLY attribute is set to true.
    *
    * @return  The boolean value of the OVERLAY attribute.
    */
   public boolean isOverlay()
   {
      return overlay;
   }

   /**
    * Setting new value for the OVERLAY flag.
    * 
    * @param   overlay
    *          The new boolean value of the OVERLAY attribute.
    */
   public void setOverlay(boolean overlay)
   {
      this.overlay = overlay;
   }

   /**
    * Getting the PAGE-BOTTOM attribute flag.
    * 
    * @return  The boolean value of the PAGE-BOTTOM attribute.
    */
   public boolean getPageBottom()
   {
      return pageBottom;
   }

   /**
    * Getting the PAGE-BOTTOM attribute flag.
    * 
    * @param   pageBottom
    *          The new boolean value of the PAGE-BOTTOM attribute.
    */
   public void setPageBottom(boolean pageBottom)
   {
      this.pageBottom = pageBottom;
   }

   /**
    * Getting the PAGE-TOP attribute flag.
    * 
    * @return  The boolean value of the PAGE-TOP attribute.
    */
   public boolean getPageTop()
   {
      return pageTop;
   }

   /**
    * Getting the PAGE-TOP attribute flag.
    * 
    * @param   pageTop
    *          The new boolean value of the PAGE-TOP attribute.
    */
   public void setPageTop(boolean pageTop)
   {
      this.pageTop = pageTop;
   }

   /**
    * Getting the value of the SIDE-LABELS flag. True means every field inside
    * the peer frame displays the label on the left side of the field,
    * otherwise - above each field.
    *
    * @return  The value of the SIDE-LABELS value.
    */
   public boolean isSideLabels()
   {
      if (instance instanceof Frame)
      {
         sideLabels = ((Frame) instance).isSideLabels();
      }
      
      return sideLabels;
   }

   /**
    * Getting the value of the SIDE-LABELS flag.
    *
    * @param   sideLabels
    *          The new value for the Progress SIDE-LABELS attribute.
    */
   public void setSideLabels(boolean sideLabels)
   {
      // This statement is intended to implicit turn on side label mode.
      // Without this the numeric FillIns will have wrong alignment
      // to the right side
      if (getColumns() > 0)
         sideLabels = true;
      
      this.sideLabels = sideLabels;
      
      if (instance instanceof Frame)
      {
         ((Frame) instance).setSideLabels(sideLabels);
      }
   }

   /**
    * Check whether the peer container is centered.
    * 
    * @return  True if the container is centered, false otherwise.
    */
   public boolean isCentered()
   {
      return centered;
   }

   /**
    * Sets the CENTERED option.
    *
    * @param centered
    *        <code>true</code> if the option is set to ON.
    */
   public void setCentered(boolean centered)
   {
      this.centered = centered;
   }

   /**
    * Check whether the frame keeps tab ordering..
    * 
    * @return  True if it does.
    */
   public boolean isKeepTabOrder()
   {
      return keepTabOrder;
   }

   /**
    * Sets the keep tab ordering option.
    *
    * @param keepTabOrder
    *        <code>true</code> if the option is set to ON.
    */
   public void setKeepTabOrder(boolean keepTabOrder)
   {
      this.keepTabOrder = keepTabOrder;
   }

   /**
    * Check whether the frame suppressed data dictionary help strings.
    * 
    * @return  True if it does.
    */
   public boolean isNoHelp()
   {
      return noHelp;
   }

   /**
    * Sets the NO-HELP option.
    *
    * @param noHelp
    *        <code>true</code> if the option is set to ON.
    */
   public void setNoHelp(boolean noHelp)
   {
      this.noHelp = noHelp;
   }

   /**
    * Check whether the frame suppressed automatic hiding.
    * 
    * @return  True if it does.
    */
   public boolean isNoHide()
   {
      return noHide;
   }

   /**
    * Sets the NO-HIDE option.
    *
    * @param noHide
    *        <code>true</code> if the option is set to ON.
    */
   public void setNoHide(boolean noHide)
   {
      this.noHide = noHide;
   }

   /**
    * Check whether the frame definition suppresses automatic hiding.
    * 
    * @return  <code>true</code> if it does.
    */
   public boolean isNoHideConfig()
   {
      return noHideConfig;
   }

   /**
    * Sets the NO-HIDE option, as configured at frame definition.
    * 
    * @param   noHideConfig
    *          <code>true</code> if the option is set to ON.
    */
   public void setNoHideConfig(boolean noHideConfig)
   {
      this.noHideConfig = noHideConfig;
   }

   /**
    * Check whether the frame suppressed labels underlining.
    * 
    * @return  True if it does.
    */
   public boolean isNoUnderline()
   {
      return noUnderline;
   }

   /**
    * Sets the NO-UNDERLINE option.
    *
    * @param noUnderline
    *        <code>true</code> if the option is set to ON.
    */
   public void setNoUnderline(boolean noUnderline)
   {
      this.noUnderline = noUnderline;
   }

   /**
    * Check whether the frame suppressed all validation from Data Dictionary.
    * 
    * @return  True if it does.
    */
   public boolean isNoValidate()
   {
      return noValidate;
   }

   /**
    * Sets the NO-VALIDATE option.
    *
    * @param noValidate
    *        <code>true</code> if the option is set to ON.
    */
   public void setNoValidate(boolean noValidate)
   {
      this.noValidate = noValidate;
   }

   /**
    * Sets the number of columns (n COLUMNS) option.
    *
    * @param columns
    *        number of columns in the frame
    */
   @Override
   public void setColumns(int columns)
   {
      super.setColumns(columns);

      if (instance instanceof Frame)
      {
         ((Frame) instance).setColumns(columns);
      }
   }

   /**
    * Gets the number of iterations to scroll in this frame.
    * 
    * @return  number of iterations to scroll at once or 0 for paging frame
    */
   public int getScroll()
   {
      return scroll;
   }

   /**
    * Sets the number of iterations to scroll in this frame.
    *
    * @param scroll
    *        number of iterations to scroll at once
    */
   public void setScroll(int scroll)
   {
      this.scroll = scroll;
   }

   /**
    * Gets the number of iterations to retain in this frame.
    * 
    * @return  number of iterations to retain at once or 0 for paging frame
    */
   public int getRetain()
   {
      return retain;
   }
   
   /**
    * Sets the number of iterations to retain in this frame.
    *
    * @param retain
    *        number of iterations to retain at once
    */
   public void setRetain(int retain)
   {
      this.retain = retain;
   }
   
   /**
    * Get current DOWN attribute value.
    * 
    * @return  current value of the attribute.
    */
   @Override
   public int getDown()
   {
      return down;
   }
   
   /**
    * Set DOWN attribute value.
    * 
    * @param   down
    *          New value for the attribute.
    */
   @Override
   public void setDown(int down)
   {
      this.down = down;
   }
   
   /**
    * Get current state of VIEW-AS DIALOG-BOX option.
    * 
    * @return  <code>true</code> if config is a DialogBox config.
    */
   public boolean isDialog()
   {
      return dialog;
   }
   
   /**
    * Set new value for the VIEW-AS DIALOG-BOX option.
    * 
    * @param   dialog
    *          New value for the option.
    */
   public void setDialog(boolean dialog)
   {
      this.dialog = dialog;
   }
   
   /**
    * Setting the Progress VISIBLE attribute.
    * <p>
    * On server-side, it will attempt to delete the frame, in case the frame is no longer visible.
    * This is done by calling {@link GenericFrame#frameDelete()}.
    * 
    * @param    visible
    *           The new value for the Progress VISIBLE attribute.
    */
   @Override
   public void setVisible(boolean visible)
   {
      super.setVisible(visible);
      
      if (!visible && frame != null)
      {
         // attempt frame delete
         frame.frameDelete();
      }
   }
   
   /**
    * Set visibility state directly.
    * <p>
    * On server-side, it will attempt to delete the frame, in case the frame is no longer visible.
    * This is done by calling {@link GenericFrame#frameDelete()}.
    * 
    * @param visible
    */
   @Override
   public void setVisibleDirect(boolean visible)
   {
      super.setVisibleDirect(visible);
      
      if (!visible && frame != null)
      {
         // attempt frame delete
         frame.frameDelete();
      }
   }
   
   /**
    * Setting the new value for HIDDEN flag.
    * <p>
    * On server-side, it will attempt to delete the frame, in case the frame is hidden.
    * This is done by calling {@link GenericFrame#frameDelete()}.
    * 
    * @param    hidden
    *           The new boolean value for the HIDDEN attribute.
    */
   @Override
   public void setHidden(boolean hidden)
   {
      super.setHidden(hidden);
      
      if (hidden && frame != null)
      {
         frame.frameDelete();
      }
   }
   
   /**
    * Replacement for the default object reading method. The latest
    * state is read from the input source.  The parent's state is
    * read first.
    * 
    * @param    in
    *           The input source from which fields will be restored.
    *
    * @throws   IOException
    *           In case of I/O errors.
    * @throws   ClassNotFoundException
    *           If payload can't be instantiated.
    */
   @Override
   public void readExternal(ObjectInput in)
   throws IOException,
          ClassNotFoundException
   {
      super.readExternal(in);
      scroll             = in.readInt();
      down               = in.readInt();
      dialog             = in.readBoolean();
      carefulPaint       = in.readBoolean();
      centered           = in.readBoolean();
      labels             = in.readBoolean();
      manualHighlight    = in.readBoolean();
      overlay            = in.readBoolean();
      pageBottom         = in.readBoolean();
      pageTop            = in.readBoolean();
      sideLabels         = in.readBoolean();
      box                = in.readBoolean();
      keepTabOrder       = in.readBoolean();
      noHelp             = in.readBoolean();
      noHide             = in.readBoolean();
      noUnderline        = in.readBoolean();
      noValidate         = in.readBoolean();
      attr_space         = in.readBoolean();
      positionChanged    = in.readBoolean();
      sizeChanged        = in.readBoolean();
   }

   /**
    * Replacement for the default object writing method. The latest
    * state is written to the output destination.  The parent's state is
    * written first.
    * 
    * @param    out
    *           The output destination to which fields will be saved.
    *
    * @throws   IOException
    *           In case of I/O errors.
    */
   @Override
   public void writeExternal(ObjectOutput out)
   throws IOException
   {
      super.writeExternal(out);
      out.writeInt(scroll);
      out.writeInt(down);
      out.writeBoolean(dialog);
      out.writeBoolean(carefulPaint);
      out.writeBoolean(centered);
      out.writeBoolean(labels);
      out.writeBoolean(manualHighlight);
      out.writeBoolean(overlay);
      out.writeBoolean(pageBottom);
      out.writeBoolean(pageTop);
      out.writeBoolean(sideLabels);
      out.writeBoolean(box);
      out.writeBoolean(keepTabOrder);
      out.writeBoolean(noHelp);
      out.writeBoolean(noHide);
      out.writeBoolean(noUnderline);
      out.writeBoolean(noValidate);
      out.writeBoolean(attr_space);
      out.writeBoolean(positionChanged);
      out.writeBoolean(sizeChanged);
   }
}
