/*
** Module   : FrameWidget.java
** Abstract : server side frame widget class (contains the configuration)
**
** Copyright (c) 2005-2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 NVS 20051026   @23491 Created initial version.
** 002 NVS 20060114   @24097 Frame is created hidden.
** 003 SIY 20060213   @24581 Fixes for the DialogBox support.
** 004 SIY 20060426   @25911 Changed signatures of get/setNextTabItem.
** 005 SIY 20060616   @27537 Added isDown(),setName() methods.
** 006 GES 20060803   @28377 Some cleanup and refactoring.
** 007 ECF 20061207   @31598 Fixed setAttrSpace(). Was calling the wrong
**                           method in FrameConfig.
** 008 CA  20090413   @41775 Fixes for FRAME:HIDDEN attribute. On init,
**                           HIDDEN flag is off. The setter must change
**                           the HIDDEN flag, not VISIBLE.
** 009 GES 20101005          Added a non-functional COM-HANDLE attribute stub.
** 010 SIY 20101006          Added moveAfterTab() stub.
** 011 SVL 20101006          #10 moved to GenericWidget.
** 012 SIY 20101007          Added setName() with different signature.
** 013 SIY 20111101          Updated imports.
** 014 CA  20130222          Added NPE protection for cases when a widget is used before is 
**                           attached to a frame (after a CREATE WIDGET statement is used).
** 015 CA  20130927          Added LegacyResource annotation, to determine resource type.
** 016 SVL 20131203          Changes caused by renaming of Nameable.get/setName to Nameable.name.
** 017 CA  20131115          Changes related to properly instantiate dynamic resources.
**                           Fixed implicit deletion of static frames.
*/

package com.goldencode.p2j.ui;

import com.goldencode.p2j.util.*;

/**
 * This is the server side frame widget class which contains the 
 * client-side configuration for this frame BUT which does NOT implement
 * any of the server side logic for frame processing.
 * <p>
 * WARNING: unlike most of the server-side widget classes, this class
 * intentionally DOES NOT push a modified screen definition to the client
 * when visibility/hidden/down changes occur.  Such changes transfer to the
 * client using alternate paths and since these are frequently changed, using
 * the screen definition for such things is not acceptable.
 */
@LegacyResource(resource = LegacyResource.FRAME)
public class FrameWidget
extends ContainerEntity
{
   /** Reference to our configuration which is pre-cast to the proper type. */
   private FrameConfig fcfg = null;
   
   /**
    * Default constructor.
    */
   public FrameWidget()
   {
      this(false);
   }
   
   /**
    * Default constructor.
    * 
    * @param    dynamic
    *           Flag indicating if this is a static or dynamic resource.
    */
   public FrameWidget(boolean dynamic)
   {
      super(dynamic, new FrameConfig(ComponentConfig.FRAME, null));
      fcfg = (FrameConfig) config;
      config.setHidden(false);
   }

   /**
    * Sets the containing frame instance.
    * 
    * @param    frame
    *           <code>GenericFrame</code> instance that contains this widget
    */
   @Override
   public void setFrame(GenericFrame frame)
   {
      super.setFrame(frame);
      
      fcfg.frame = frame;
   }
   
   /**
    * Set NAME attribute.
    * 
    * @param   name
    *          New value for the attribute.
    */
   public void name(String name)
   {
      fcfg.setName(name);
   }
   
   /**
    * Set NAME attribute.
    * 
    * @param   name
    *          New value for the attribute.
    */
   public void name(character name)
   {
      name(name.isUnknown() ? "" : name.toStringMessage());
   }
   
   /**
    * Gets the NEXT-TAB-ITEM writable attribute.
    *
    * @return   The current value of NEXT-TAB-ITEM attribute.
    */
   public GenericWidget getNextTabItem()
   {
      // TODO: implement this!
      throw new RuntimeException("NEXT-TAB-ITEM is not done yet!");
   }

   /**
    * Sets the NEXT-TAB-ITEM writable attribute.
    *
    * @param    n
    *           The new value for the NEXT-TAB-ITEM attribute.
    */
   public void setNextTabItem(GenericWidget n)
   {
      // TODO: implement this!
      throw new RuntimeException("NEXT-TAB-ITEM is not done yet!");
   }

   /**
    * Gets the ATTR-SPACE writable attribute.
    *
    * @return   The current value of the ATTR-SPACE attribute.
    */
   public logical isAttrSpace()
   {
      return new logical(fcfg.isAttrSpace());
   }

   /**
    * Sets the ATTR-SPACE writable attribute.
    *
    * @param    space
    *           The new value for the ATTR-SPACE attribute.
    */
   public void setAttrSpace(logical space)
   {
      setAttrSpace(space.booleanValue());
   }

   /**
    * Sets the ATTR-SPACE writable attribute.
    *
    * @param    space
    *           The new value for the ATTR-SPACE attribute.
    */
   public void setAttrSpace(boolean space)
   {
      fcfg.setAttrSpace(space);
      pushScreenDefinition();
   }
   
   /**
    * Sets the dialog box flag.
    *
    * @param dialogBox
    *        new value for the dialog box flag
    */
   public void setDialogBox(boolean dialogBox)
   {
      fcfg.setDialog(dialogBox);
   }
   
   /**
    * Gets the dialog box flag.
    *
    * @return <code>true</code> if this frame is a dialog box.
    */
   public boolean isDialogBox()
   {
      return fcfg.isDialog();
   }

   /**
    * Sets the down frame size.
    *
    * @param    down
    *           The new number of iterations for this down frame.
    */
   public void setDown(int down)
   {
      fcfg.setDown(down);
   }
   
   /**
    * Gets the down frame size.
    *
    * @return   The number of data iterations for the frame (1 if the
    *           frame is not a down frame).
    */
   public int getDown()
   {
      return fcfg.getDown();
   }

   /**
    * Return <code>true</code> if frame is DOWN frame.
    * 
    * @return   <code>true</code> if frame is DOWN frame.
    */
   public boolean isDown()
   {
      return (getDown() != 1);
   }
   
   /**
    * Sets the new value for HIDDEN flag.
    * 
    * @param   hidden
    *          The new boolean value for the HIDDEN attribute.
    */
   public void setHidden(boolean hidden)
   {
      config.setHidden(hidden);
   }

   /**
    * Sets the Progress VISIBLE attribute.
    * 
    * @param   visible
    *          The new value for the Progress VISIBLE attribute.
    */
   public void setVisible(boolean visible)
   {
      config.setVisible(visible);
   }
   
   /**
    * Gets the Progress COM-HANDLE attribute.
    * <p>
    * This is a non-functional stub for now.
    * 
    * @return   The value for the COM-HANDLE attribute.
    */
   public comhandle getComHandle()
   {
      return new comhandle();
   }
   
   /**
    * Hides this widget.
    * <p>
    * This method corresponds to the Progress statement: HIDE widget
    * 
    * @param   noPause
    *          <code>true</code> causes omission of any pauses which would
    *          be necessary otherwise.
    */
   @Override
   public void hide(boolean noPause)
   {
      frame.hide(noPause);
   }
   
   /**
    * Check if the resource may be implicitly deleted. 
    * 
    * @return   <code>true</code> if the INSTANTIATING-PROCEDURE attribute is unknown (i.e. the 
    *           instantiating procedure has been deleted).
    */
   protected boolean implicitDeletion()
   {
      // shared frames can be declared only in external procedure; they are always marked as for
      // implicit deletion; resourceDelete will take care to delete only the first reference
      return frame.isShared()                      || 
             !TransactionManager.isExternalBlock() || 
             instantiatingProcedure().isUnknown();
   }

   /**
    * Delete the resource.
    * 
    * @return   <code>true</code> if the resource was deleted.
    */
   @Override
   protected boolean resourceDelete()
   {
      if (!canDelete())
      {
         return false;
      }
      
      // check frame special cases
      if (!_dynamic()        && 
          config.isVisible() &&  
          !frame.isShared()  &&
          !(frame.isRedirected() || frame.isOutRedirected()))
      {
         // implicit static frame deletion is possible only if is invisible
         return false;
      }
      
      if (frame.isShared() && frame.getScopeNesting() > 0)
      {
         // shared frames can't be deleted until all are out of scope.
         return false;
      }
      
      deleted = true;
      
      // remove the frame, etc
      frame.frameCleanup();
      
      return true;
   }

   /**
    * Set the instantiating procedure for this resource.
    * <p>
    * This will set the instantiating procedure only if is not shared or this is the procedure
    * where the frame is registered as shared.
    *  
    * @param    hproc
    *           The procedure where the resource was instantiated.
    */
   @Override
   protected void setInstantiatingProcedure(handle hproc)
   {
      if (!frame.isShared() || frame.getScopeNesting() == 0)
      {
         this.instProc.assign(hproc);
      }
   }

   /**
    * Force new instantiating procedure for this frame.
    * 
    * @param    hproc
    *           The procedure where the resource was instantiated. For shared frames, this must be
    *           the procedure where the frame was registered as shared.
    */
   protected void setSharedInstantiatingProcedure(handle hproc)
   {
      this.instProc.assign(hproc);
   }
}
