/*
** Module   : AbstractWidget.java
** Abstract : Base class for widgets.
**
** Copyright (c) 2010-2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 SIY 20100629 Created initial version
** 002 SIY 20100804 Changes in imports, some refactorings.
** 003 SIY 20100822 Added base implementation of new interface methods config() and isFocusable().
** 004 SIY 20101122 Switching from KeyEvent to KeyInput.
** 005 SIY 20111115 Added physical widget location.
** 006 CA  20131121 The hidden/visible state changes must be collected when the widget's attribute
**                  is actually set.
*/

package com.goldencode.p2j.ui.client.widget;

import java.lang.ref.*;
import java.util.*;

import com.goldencode.p2j.ui.chui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.event.listener.*;

/**
 * Base widget class. 
 */
public abstract class AbstractWidget 
implements Widget
{
   /** Widget enabled flag. */
   private boolean enabled = true;
   
   /** Widget hidden flag. */
   private boolean hidden = false;
   
   /** Widget highlighted flag. */
   private boolean highlighted = false;
   
   /** Widget visible flag. */
   private boolean visible = true;

   /** Storage for focus listeners. */
   private ArrayList<FocusListener> focusListeners = new ArrayList<FocusListener>();
   
   /** Storage for key listeners. */
   private ArrayList<KeyListener> keyListeners = new ArrayList<KeyListener>();

   /** Widget name. */
   private String name = "";
   
   /** Widget origin (upper left corner) relative to container expressed in
    * character cell units. */
   private Point origin = new Point(0, 0);

   /** Widget physical origin (upper left corner) relative to container. 
       expressed in units native to UI type. */
   private Point physicalOrigin = new Point(0, 0);
   
   /** Reference to parent container. */
   private WeakReference<Container> parent = null;

   /** Reference to screen. */
   private OutputManager screen;

   /**
    * Constructor
    */
   public AbstractWidget()
   {
      screen = OutputManager.instance();
   }

   /**
    * Get access to {@link OutputManager} instance.
    * 
    * @return  reference to {@link OutputManager}.
    */
   @Override
   public OutputManager screen() 
   {
      return screen;
   }

   /**
    * Add focus listener.
    * 
    * @param   listener
    *          Listener to add.
    */
   @Override
   public void addFocusListener(FocusListener listener)
   {
      focusListeners.add(listener);
   }

   /**
    * Add key listener.
    * 
    * @param   listener
    *          Listener to add.
    */
   @Override
   public void addKeyListener(KeyListener listener)
   {
      keyListeners.add(listener);
   }

   /**
    * Get ancestor window for this widget.
    * 
    * @return  ancestor window.
    */
   @Override
   public TitledWindow ancestor()
   {
      if (this instanceof TitledWindow)
         return (TitledWindow) this;

      Container ancestor;
      Container nextancestor;

      for (ancestor = parent(); !(ancestor instanceof TitledWindow); ancestor = nextancestor)
      {
         if (ancestor == null)
            return null;

         nextancestor = ancestor.parent();
         
         if (nextancestor == null)
            return null;
      }
      
      return (TitledWindow) ancestor;
   }

   /**
    * Get widget bounding rectangle.
    * 
    * @return  bounding rectangle.
    */
   @Override
   public Rectangle bounds()
   {
      return new Rectangle(origin, dimension());
   }

   /**
    * Get widget dimension.
    * 
    * @return  widget dimension.
    */
   @Override
   public abstract Dimension dimension();

   /**
    * Check if widget is displayed.
    * 
    * @return  <code>true</code> if widget is displayed.
    */
   @Override
   public boolean isDisplayed()
   {
      return (parent() == null) ? false : parent().isDisplayed();
   }

   /**
    * Draw widget.
    */
   @Override
   public abstract void draw();

   /**
    * Draw caret.
    */
   @Override
   public void drawCaret()
   {
   }

   /**
    * Check if widget is enabled.
    * 
    * @return  <code>true</code> if widget is enabled.
    */
   @Override
   public boolean isEnabled()
   {
      return enabled;
   }

   /**
    * Check if widget is focus traversable.
    * 
    * @return  <code>true</code> if widget is focus traversable.
    */
   @Override
   public boolean focusTraversable()
   {
      return (enabled && visible && !hidden);
   }

   /**
    * Force widget visibility change without associated repainting.
    * 
    * @param   visibility
    *          New visibility state.
    */
   @Override
   public void forceVisibility(boolean visibility)
   {
      ThinClient.getInstance().collectFlag(this, visibility, true);
      
      visible  = visibility;
   }

   /**
    * Check if widget has focus.
    * 
    * @return  <code>true</code> if widget has focus.
    */
   @Override
   public boolean hasFocus()
   {
      return (parent() == null || parent().focus() != this) ? 
                                                         false : 
                                                         parent().hasFocus();
   }

   /**
    * Get widget height.
    * 
    * @return  widget height.
    */
   @Override
   public abstract int height();

   /**
    * Get current hidden state.
    * 
    * @return   hidden state.
    */
   @Override
   public boolean hidden()
   {
      return hidden;
   }

   /**
    * Hide widget.
    */
   @Override
   public void hide()
   {
      if (visible)
      {
         ThinClient.getInstance().collectFlag(this, false, true);
         
         visible = false;
         repaint();
      }
   }

   /**
    * Get current highlighted state.
    * 
    * @return   highlighted state.
    */
   @Override
   public boolean highlighted()
   {
      return highlighted;
   }

   /**
    * Get widget location relative to container.
    * 
    * @return  relative widget location.
    */
   @Override
   public Point location()
   {
      return new Point(origin);
   }

   /**
    * Get widget physical location relative to container.
    * 
    * @return  relative widget location.
    */
   @Override
   public Point physicalLocation()
   {
      return new Point(physicalOrigin);
   }
   
   /**
    * Get widget minimum size.
    * 
    * @return  widget minimum size.
    */
   @Override
   public abstract Dimension minimumSize();

   /**
    * Get widget name.
    * 
    * @return  widget name.
    */
   @Override
   public String name()
   {
      return name;
   }

   /**
    * Get widget parent.
    * 
    * @return  widget parent.
    */
   @Override
   public Container parent()
   {
      return (parent == null) ? null : parent.get();
   }

   /**
    * Process focus event.
    * 
    * @param   event
    *          Event to process.
    */
   @Override
   public void processFocusEvent(FocusEvent event)
   {
      for (FocusListener listener : focusListeners)
      {
         event.dispatch(listener);
      }
   }

   /**
    * Process key event.
    * 
    * @param   event
    *          Key event to process.
    */
   @Override
   public void processKeyEvent(KeyInput event)
   {
      processKeyListeners(event);
   }

   /**
    * Process key listeners.
    * 
    * @param    event
    *           Keyboard event.
    */
   @Override
   public void processKeyListeners(KeyInput event)
   {
      for (KeyListener listener : keyListeners)
      {
         event.dispatch(listener);

         if (event.isConsumed())
            break;
      }
   }

   /**
    * Request widget repainting.
    */
   @Override
   public void repaint()
   {
      if (!isDisplayed())
         return;

      repaint(bounds());
   }

   /**
    * Request widget repainting.
    * 
    * Only the given update rectangle will be repainted (note that it is
    * clipped by all parent containers). No event is posted if the widget
    * itself or any of the parent containers are hidden.
    * 
    * @param    ur
    *           The update rectangle (relative to the origin of its parent
    *           Container) which will be repainted. Note that it is clipped by
    *           all parent containers.
    */
   @Override
   public void repaint(Rectangle ur)
   {
      if (!isDisplayed())
         return;

      Widget cm = this;
      Container ct = parent();

      while (ct != null)
      {
         if (!ct.isVisible())
         {
            return;
         }

         if (ct.clipping())
         {
            ur = ur.intersection(ct.getClipRect());

            if (ur == null || ur.empty())
            {
               return;
            }
         }

         cm = ct;
         ct = cm.parent();
         ur = ur.translate(cm.location());
      }

      EventManager.postEvent(new PaintEvent(this, ur));
   }

   /**
    * Request focus.
    */
   @Override
   public void requestFocus()
   {
      Widget currentFocus = ancestor().currentFocus();
      
      if (currentFocus == this)
         return;
      
      if (currentFocus != null)
         EventManager.postEvent(new FocusEvent(currentFocus, EventType.FOCUS_LOST));

      EventManager.postEvent(new FocusEvent(this, EventType.FOCUS_GAINED));

      parent().setFocus(this);
      repaint();
   }

   /**
    * Request screen synchronization.  
    */
   @Override
   public void requestSync()
   {
      EventManager.postEvent(new SyncEvent(this));
   }

   /**
    * Get absolute location of widget in screen coordinates.
    * 
    * @return  absolute widget location.
    */
   @Override
   public Point screenLocation()
   {
      if (parent() == null) {
         throw new IllegalStateException("widget is not in the container");
      }

      return parent().screenLocation().translate(origin);
   }

   /**
    * Enable/disable widget.
    * 
    * @param   enabled
    */
   @Override
   public void setEnabled(boolean enabled)
   {
      this.enabled = enabled;

      repaint();
   }

   /**
    * Tracks the hidden state of the widget.
    * 
    * @param    hidden
    *           New hidden state.
    */
   @Override
   public void setHidden(boolean hidden)
   {
      ThinClient.getInstance().collectFlag(this, hidden, false);

      this.hidden = hidden;
   }

   /**
    * Tracks the highlighted state of the widget.
    * 
    * @param    highlighted
    *           New highlighted state.
    */
   @Override
   public void setHighlighted(boolean highlighted)
   {
      this.highlighted = highlighted;
   }

   /**
    * Set widget relative location.
    * 
    * @param   x
    *          Widget X.
    * @param   y
    *          Widget Y.
    */
   @Override
   public void setLocation(int x, int y)
   {
      if (visible)
         repaint(); 
      
      origin.x = x;
      origin.y = y;
      
      if (visible)
         repaint();
   }

   /**
    * Set widget relative location.
    * 
    * @param   origin
    *          Location to set.
    */
   @Override
   public void setLocation(Point origin)
   {
      setLocation(origin.x, origin.y);
   }

   /**
    * Set widget relative physical location.
    * 
    * @param   x
    *          Widget physical X.
    * @param   y
    *          Widget physical Y.
    */
   @Override
   public void setPhysicalLocation(int x, int y)
   {
      // note that these values not checked nor processed in any way,
      // widget just holds them as a container
      physicalOrigin.x = x;
      physicalOrigin.y = y;
   }
   
   /**
    * Set widget name.
    * 
    * @param   name
    *          New widget name.
    */
   @Override
   public void setName(String name)
   {
      this.name = name;
   }

   /**
    * Set widget parent.
    * 
    * @param   container
    *          Parent to set.
    */
   @Override
   public void setParent(Container container)
   {
      parent = (container == null) ? 
               null : 
               new WeakReference<Container>(container);
   }

   /**
    * Hide/show widget.
    * 
    * @param   visible
    *          New visible state.
    */
   @Override
   public void setVisible(boolean visible)
   {
      if (visible)
         show();
      else
         hide();
   }

   /**
    * Show widget.
    */
   @Override
   public void show()
   {
      if (!visible)
      {
         ThinClient.getInstance().collectFlag(this, true, true);

         visible = true;
         repaint();
      }
   }

   /**
    * Check if widget is totally obscured.
    * 
    * @return  <code>true</code> if widget is totally obscured.
    */
   @Override
   public boolean isTotallyObscured()
   {
      return WindowManager.checkRectangle(bounds(), ancestor());
   }

   /**
    * Check if widget is visible.
    * 
    * @return  <code>true</code> if widget is visible.
    */
   @Override
   public boolean isVisible()
   {
      return visible;
   }

   /**
    * Get widget width.
    * 
    * @return  widget width.
    */
   @Override
   public abstract int width();

   /**
    * Disable widget.
    */
   protected void disable()
   {
      enabled = false;
   }

   /**
    * Enable widget.
    */
   protected void enable()
   {
      enabled  = true;
   }

   /**
    * Process event and dispatch it to specific method.
    * 
    * @param   evt
    *          Event to process.
    */
   @Override
   public void processEvent(Event evt)
   {
      if (evt instanceof KeyInput)
      {
         Widget ancestor = ancestor();

         if (ancestor != null)
            ancestor.processKeyEvent((KeyInput) evt);
      }
      else if (evt instanceof FocusEvent)
         processFocusEvent((FocusEvent) evt);
   }

   /**
    * By default widget has no config.
    */
   @Override
   public <T extends WidgetConfig> T config()
   {
      return null;
   }

   /**
    * Checks whether this widget can receive focus generally.
    * 
    * @return  <code>true</code> if this widget can be focused under
    *          certain conditions. <code>false</code> if, and only if, this
    *          widget can never receive focus.    
    */
   @Override
   public boolean isFocusable()
   {
      return true;
   }
}
