/*
** Module   : HandleResource.java
** Abstract : Implementation of common attributes and APIs for all resources. 
**
** Copyright (c) 2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20130221 Created initial version.
** 002 OM  20130304 Refactored isValid and isUnknown of WrappedResource to valid and unknown.
** 003 EVK 20130903 Added method type().
** 004 CA  20130927 All resources are known by default; special resources which need to be unknown
**                  will override the unknown() API. The resource type is determined based on the
**                  LegacyResource annotation: an annotation at the class has precedence over an
**                  annotation at the interface.
** 005 CA  20131113 Added INSTANTIATING-PROCEDURE support.
**                  Changes related to properly instantiate dynamic resources.
*/

package com.goldencode.p2j.util;

import java.util.*;

import com.goldencode.proxy.*;

/**
 * This acts as a superclass which implements:
 * <table>
 * <tr>
 *    <th>attribute or base method</th>
 *    <th>API</th>
 * </tr>
 * <tr>
 *    <td>TYPE</td>
 *    <td>{@link #getResourceType()} and {@link #type()}</td>
 * </tr>
 * <tr>
 *    <td>INSTANTIATING-PROCEDURE</td>
 *    <td>{@link #instantiatingProcedure()}</td>
 * </tr>
 * <tr>
 *    <td>Unknown status</td>
 *    <td>{@link #unknown()}</td>
 * </tr>
 * <tr>
 *    <td>Read only field error</td>
 *    <td>{@link #readOnlyError(String)}</td>
 * </tr>
 * <tr>
 *    <td>Invalid attribute access</td>
 *    <td>{@link #invalidAttribute(String)}</td>
 * </tr>
 * </tr>
 * </table>
 * <p>
 * The abstract class also provides setters for the implemented attributes 
 * which are not defined inside the interfaces to be used for initializing of
 * those attributes
 */
public abstract class HandleResource
implements CommonHandle,
           InstantiatingProcedure
{
   /** Handle referencing the instantiating procedure. */
   protected final handle instProc = new handle();

   /** Variable for holding the type attribute */
   private String type = null;
   
   /**
    * Default c'tor.
    */
   public HandleResource()
   {
      this(false);
   }
   
   /**
    * Initialize this resource by determining its type.
    * <p>
    * The resource's type is determined from a {@link LegacyResource} annotation, which is 
    * searched in these locations, in this order:
    * <ol>
    * <li>The class implementing the resource in P2J is checked for an annotation. If found, this
    *     is used to determine the resource type.</li>
    * <li>The entire interface hierarchy is checked. There should be only one interface with a
    *     {@link LegacyResource} annotation; if there is more than annotated interface, then n
    *     {@link IllegalStateException} is thrown.</li>
    * </ol>
    * 
    * @param    dynamic
    *           Flag indicating if this is a static or dynamic resource.
    * 
    * @throws   IllegalStateException
    *           If the resource type could not be determined from a {@link LegacyResource} 
    *           annotation, after checking the interface hierarchy, or more than one interface
    *           is annotated.
    */
   protected HandleResource(boolean dynamic)
   {
      // check the class first
      Class<?> cls = getClass();
      LegacyResource lr = cls.getAnnotation(LegacyResource.class);
      if (lr != null)
      {
         type = lr.resource();
      }
      else
      {
         // else, determine it from the interface. the annotated interface in most cases is a 
         // directly implemented interface, but ensure we reach all interfaces, to cover all
         // bases

         Deque<Class<?>> ifcs = new ArrayDeque<Class<?>>();
         // add all the interfaces from the extended classes first
         Class<?> superCls = cls;
         while (superCls != null)
         {
            ifcs.addAll(Arrays.asList(superCls.getInterfaces()));
            
            superCls = superCls.getSuperclass();
         }
         
         Set<Class<?>> checked = new HashSet<Class<?>>();
         
         while (!ifcs.isEmpty())
         {
            Class<?> ifc = ifcs.remove();
            
            if (checked.contains(ifc))
               continue;
            
            // keep track of which interface we've visited, as the same interface may appear more
            // than once in the hierarchy
            checked.add(ifc);
            
            lr = ifc.getAnnotation(LegacyResource.class);
            if (lr != null)
            {
               if (type != null)
               {
                  // if more than one interface is annotated, then we are ambiguous
                  final String msg = 
                     "More than one LegacyResource annotation in the implemented interfaces " +
                     "for class %s.";
                  throw new IllegalStateException(String.format(msg, cls.getName()));
               }

               type = lr.resource();
            }
            
            ifcs.addAll(Arrays.asList(ifc.getInterfaces()));
         }
      }

      if (type == null)
      {
         final String msg = "Could not determine the legacy resource type for class %s.";
         throw new IllegalStateException(String.format(msg, cls.getName()));
      }
      
      // this will ensure the INSTANTIATING-PROCEDURE attribute is setup accordingly and also that
      // static resources will be registered with the correct top-level block
      if (!ProxyFactory.isProxyClass(this.getClass()))
      {
         // for proxy cases, explicit registration is needed
         ProcedureManager.processResource(this, dynamic);
      }
   }
   
   /**
    * Get the type of this handle.
    * 
    * @return   See above.
    */
   public character getResourceType()
   {
      return new character(type);
   }

   /**
    * A newly-created resource is always not-unknown. If implementations need to return something
    * different, then override this method.
    * 
    * @return   always <code>false</code>
    */
   @Override
   public boolean unknown()
   {
      return false;
   }
   
   /**
    * To be used only in cases when a setter for a read-only attribute or a not-supported  
    * attribute is forced to be implemented by the resource. It displays appropriate error 
    * message.
    * 
    * @param    attr
    *           The read-only attribute.
    */
   public void readOnlyError(String attr)
   {
      handle.readOnlyError(new handle(this), attr);
   }

   /**
    * Get the type of this handle as type of String.
    *
    * @return   See above.
    */
   public String type()
   {
      return type;
   }
   
   /**
    * Getter for the INSTANTIATING-PROCEDURE attribute.
    * <p>
    * The returned handle will refer a valid external procedure (persistent or not) or will be
    * unknown, in case its instantiating procedure was persistent and has been deleted.
    * 
    * @return   See above.
    */
   @Override
   public handle instantiatingProcedure()
   {
      return ProcedureManager.resolveInstantiatingProcedure(instProc);
   }
   
   /**
    * To be used only in cases a getter for a not-supported attribute is forced to be implemented
    * by the resource. It displays appropriate error message.
    * 
    * @param    attr
    *           The not-supported attribute attribute.
    */
   protected void invalidAttribute(String attr)
   {
      handle.invalidAttribute(attr, new handle(this));
   }
   
   /**
    * Check if the resource may be implicitly deleted. 
    * 
    * @return   <code>true</code> if the INSTANTIATING-PROCEDURE attribute is unknown (i.e. the 
    *           instantiating procedure has been deleted) and we are during finalizable or
    *           scopeable notification.
    */
   protected boolean implicitDeletion()
   {
      return (!TransactionManager.isExternalBlock() || instantiatingProcedure().isUnknown()) &&
             ProcedureManager.inNotifications();
   }

   /**
    * Set the instantiating procedure for this resource.
    * 
    * @param    hproc
    *           The procedure where the resource was instantiated.
    */
   protected void setInstantiatingProcedure(handle hproc)
   {
      this.instProc.assign(hproc);
   }
}
